#include <Arduino.h>
#include "LineBuffer.h"

// LineBuffer のコンストラクタ
LineBuffer::LineBuffer(int _length) {
  this->length = 0;
  this->data = NULL;
  this->allocate(_length);
  this->count = 0;
}

// LineBuffer のデストラクタ
LineBuffer::~LineBuffer() {
  this->deallocate();
}

// バッファ領域の確保
void LineBuffer::allocate(int _length) {
  this->data = new char[_length];
  this->length = this->data ? _length : 0;
}

// バッファ領域の解放
void LineBuffer::deallocate() {
  delete[] this->data;
}

// データの有無を確認。1 バイトでもあれば 1 を、なければ 0 を返却
bool LineBuffer::available() {
  return this->count > 0;
}

// データを全消去
void LineBuffer::clear() {
  this->count = 0;
}

// データ 1 バイトを格納。成功時 0 を、失敗時非 0 を返却
int LineBuffer::put(char _c) {
  if (this->count == this->length) {
    return 1;
  } else {
    this->data[this->count++] = _c;
    return 0;
  }
}

// 現在データサイズを取得
int LineBuffer::getCount() {
  return this->count;
}

// バッファ領域へのポインタを取得
char* LineBuffer::getData() {
  return this->data;
}

