#include <Arduino.h>
#include <ctype.h>
#include "Common.h"
#include "Parser.h"
#include "ctypex.h"

Parser::Parser(int _length) {
  this->buffer = new LineBuffer(_length);
  this->bufferCompleted = false;
  this->lineCount = 0;
}

Parser::~Parser() {
  delete this->buffer;
}

void Parser::clear() {
  this->buffer->clear();
  this->bufferCompleted = false;
}

void Parser::setLineCount(int _n) {
  this->lineCount = _n;
}

int Parser::put(char _c) {
  if (!this->bufferCompleted) {
    if (_c == '\r') {
      this->bufferCompleted = true;
      this->buffer->put('\n');
    } else if (_c != '\n') {
      this->buffer->put(toupper(_c));
    }
  }
  return this->bufferCompleted;
}

bool Parser::parseIdentifier(Token& _t, char* _s, int _count) {
  bool valid = true;
  Token result;
  int n, x;
  result.type = TokenType::UNDEFINED;
  result.numErr = NumericError::NO_ERROR;
  result.value = 0;

  if (valid) {
    // 識別子の特定
    switch (_count) {
      case 3:
        n = _s[1] - 'A';
        x = _s[2] - '0';
        if (strncmp("TP", _s, 2) == 0) {
          if (x >= 0 && x < 8) {
            result.type = (TokenType) ((int) TokenType::BI_TIMER_POSEDGE0 + x);
          } else valid = false;
        } else if (strncmp("TN", _s, 2) == 0) {
          if (x >= 0 && x < 8) {
            result.type = (TokenType) ((int) TokenType::BI_TIMER_NEGEDGE0 + x);
          } else valid = false;
        } else if (strncmp("TL", _s, 2) == 0) {
          if (x >= 0 && x < 8) {
            result.type = (TokenType) ((int) TokenType::BI_TIMER_LEVELTRIG0 + x);
          } else valid = false;
        } else if (strncmp("TR", _s, 2) == 0) {
          if (x >= 0 && x < 8) {
            result.type = (TokenType) ((int) TokenType::BI_TIMER_RESET0 + x);
          } else valid = false;
        } else if (strncmp("D", _s, 1) == 0) {
          if (n >= 0 && n < 8 && x >= 0 && x < 4) {
            result.type = (TokenType)
                          ((int) TokenType::BI_ANDGATE_INA0 + (n << 2) + x);
          } else if (_s[1] == 'Y') {
            if (x >= 0 && x < 4) {
              result.type = (TokenType) ((int) TokenType::BO_ANDGATE_OUT0 + x);
            } else {
              valid = false;
            }
          } else {
            valid = false;
          }
        } else if (strncmp("W", _s, 1) == 0) {
          if (n >= 0 && n < 8 && x >= 0 && x < 4) {
            result.type = (TokenType)
                          ((int) TokenType::BI_ORGATE_INA0 + (n << 2) + x);
          } else if (_s[1] == 'Y') {
            if (x >= 0 && x < 4) {
              result.type = (TokenType) ((int) TokenType::BO_ORGATE_OUT0 + x);
            } else {
              valid = false;
            }
          } else {
            valid = false;
          }
        } else if (strncmp("NX", _s, 2) == 0) {
          if (x >= 0 && x < 8) {
            result.type = (TokenType) ((int) TokenType::BI_NOTGATE_IN0 + x);
          } else valid = false;
        } else if (strncmp("RY", _s, 2) == 0) {
          if (x >= 0 && x < 4) {
            result.type = (TokenType) ((int) TokenType::BI_RELAY0 + x);
          } else valid = false;
        } else if (strncmp("JL", _s, 2) == 0) {
          if (x >= 0 && x < 6) {
            result.type = (TokenType) ((int) TokenType::II_JP_MUXL0 + x);
          } else valid = false;
        } else if (strncmp("JH", _s, 2) == 0) {
          if (x >= 0 && x < 6) {
            result.type = (TokenType) ((int) TokenType::II_JP_MUXH0 + x);
          } else valid = false;
        } else if (strncmp("TX", _s, 2) == 0) {
          if (x >= 0 && x < 8) {
            result.type = (TokenType) ((int) TokenType::II_TIMER_CNT0 + x);
          } else valid = false;
        } else if (strncmp("TM", _s, 2) == 0) {
          if (x >= 0 && x < 8) {
            result.type = (TokenType) ((int) TokenType::II_TIMER_MULT0 + x);
          } else valid = false;
        } else if (strncmp("A", _s, 1) == 0) {
          if (n >= 0 && n < 2 && x >= 0 && x < 4) {
            result.type = (TokenType)
                          ((int) TokenType::II_ADDER_INA0 + (n << 2) + x);
          } else if (_s[1] == 'Y') {
            if (x >= 0 && x < 4) {
              result.type = (TokenType) ((int) TokenType::IO_ADDER_OUT0 + x);
            } else {
              valid = false;
            }
          } else {
            valid = false;
          }
        } else if (strncmp("M", _s, 1) == 0) {
          if (n >= 0 && n < 2 && x >= 0 && x < 4) {
            result.type = (TokenType)
                          ((int) TokenType::II_MULTIPLIER_INA0 + (n << 2) + x);
          } else if (_s[1] == 'Y') {
            if (x >= 0 && x < 4) {
              result.type = (TokenType) ((int) TokenType::IO_MULTIPLIER_OUT0 + x);
            } else {
              valid = false;
            }
          } else {
            valid = false;
          }
        } else if (strncmp("VM", _s, 2) == 0) {
          if (x >= 0 && x < 4) {
            result.type = (TokenType) ((int) TokenType::II_VR_MAX0 + x);
          } else valid = false;
        } else if (strncmp("PP", _s, 2) == 0) {
          if (_s[2] == 'D') {
            result.type = TokenType::BO_CONST_DELAY;
          } else if (_s[2] == 'C') {
            result.type = TokenType::BO_CONST_ALWAYS;
          } else valid = false;
        } else if (strncmp("PH", _s, 2) == 0) {
          if (x >= 0 && x < 4) {
            result.type = (TokenType) ((int) TokenType::BO_PHOTO0 + x);
          } else valid = false;
        } else if (strncmp("JP", _s, 2) == 0) {
          if (x >= 0 && x < 6) {
            result.type = (TokenType) ((int) TokenType::BO_JP_LEVEL0 + x);
          } else valid = false;
        } else if (strncmp("TY", _s, 2) == 0) {
          if (x >= 0 && x < 8) {
            result.type = (TokenType) ((int) TokenType::BO_TIMER_OUT0 + x);
          } else valid = false;
        } else if (strncmp("TD", _s, 2) == 0) {
          if (x >= 0 && x < 8) {
            result.type = (TokenType) ((int) TokenType::BO_TIMER_DELAY0 + x);
          } else valid = false;
        } else if (strncmp("NY", _s, 2) == 0) {
          if (x >= 0 && x < 8) {
            result.type = (TokenType) ((int) TokenType::BO_NOTGATE_OUT0 + x);
          } else valid = false;
        } else if (strncmp("JY", _s, 2) == 0) {
          if (x >= 0 && x < 6) {
            result.type = (TokenType) ((int) TokenType::IO_JP_MUXOUT0 + x);
          } else valid = false;
        } else if (strncmp("VR", _s, 2) == 0) {
          if (x >= 0 && x < 4) {
            result.type = (TokenType) ((int) TokenType::IO_VR_OUT0 + x);
          } else valid = false;
        } else if (strncmp("RUN", _s, 3) == 0) {
          result.type = TokenType::CMD_RUN;
        } else {
          valid = false;
        }
        break;
      case 4:
        if (strncmp("STOP", _s, 4) == 0) {
          result.type = TokenType::CMD_STOP;
        } else if (strncmp("ENDB", _s, 4) == 0) {
          result.type = TokenType::CMD_ENDBLOCK;
        } else if (strncmp("DUMP", _s, 4) == 0) {
          result.type = TokenType::CMD_DUMP;
        } else if (strncmp("MAJV", _s, 4) == 0) {
          result.type = TokenType::NI_MAJOR_VERSION;
        } else if (strncmp("MINV", _s, 4) == 0) {
          result.type = TokenType::NI_MINOR_VERSION;
        } else {
          valid = false;
        }
        break;
      case 6:
        if (strncmp("BLOCK", _s, 5) == 0) {
          x = _s[5] - '0';
          if (x >= 0 && x < NUM_BLOCKS && x < 10) {
            result.type = (TokenType) ((int) TokenType::CMD_BLOCK0 + x);
          } else valid = false;
        } else {
          valid = false;
        }
        break;
      default:
        valid = false;
        break;
    }
  }
  if (!valid) {
    result.type = TokenType::UNDEFINED;
    Serial.print(F("Error[PI01] at line "));
    Serial.println(this->lineCount);
    Serial.print(F("Undefined identifier \""));
    for (int i = 0; i < _count; i++) {
      Serial.print(_s[i]);
    }
    Serial.println("\"");
  }
  _t = result;
  return valid;
}

bool Parser::parseNumber(Token& _t, char* _s, int _count) {
  bool valid = true;
  Token result;
  bool negative = false;  // false ならば正またはゼロ、true ならば負
  int value = 0;
  long newValue;
  char c;
  result.type = TokenType::IMMEDIATE;
  result.numErr = NumericError::NO_ERROR;
  result.value = 0;

  for (int i = 0; i < _count && valid; i++) {
    c = *_s++;
    if (isdigit(c)) {
      // 数字
      newValue = 10L * value + (c - '0');
      if ((!negative && newValue > 32767L)
          || (negative && newValue > 32768L)) {
        // オーバフロー
        valid = false;
        result.numErr = NumericError::OVERFLOW;
        Serial.print(F("Error[PN01] at line "));
        Serial.println(this->lineCount);
        Serial.println(F("Number overflow"));
      } else {
        value = (int) newValue;
      }
    } else if (c == '+' || c == '-') {
      // 符号: 先頭の + もしくは - のみ受理。先頭以外はエラー
      if (i == 0) {
        negative = c == '-';
      } else {
        valid = false;
        result.numErr = NumericError::BAD_SIGN;
        Serial.print(F("Error[PN02] at line "));
        Serial.println(this->lineCount);
        Serial.println(F("Bad sign in number"));
      }
    } else {
      // それ以外の文字はエラー
      valid = false;
      result.numErr = NumericError::ILLEGAL_CHAR;
    }
  }
  if (valid) {
    result.value = negative ? -value : value;
  }
  _t = result;
  return valid;
}

bool Parser::parse(Token& _t0, Token& _t1) {
  bool valid = true;

  if (this->bufferCompleted) {
    int count = this->buffer->getCount();
    char *data = this->buffer->getData();
    char c;
    char field[FIELD_LENGTH];
    int fieldId = 0;
    int fieldCount = 0;
    enum {
      INITIAL = 0,
      IDENTIFIER = 1,
      NUMBER = 2,
      COMMA = 3,
      COMPLETED = 4,
      ERROR = 5,
    } tokenStatus = INITIAL, nextTokenStatus = INITIAL;

    // トークン結果初期化
    _t0.type = TokenType::EMPTY;
    _t1.type = TokenType::EMPTY;
    bool commaInPlace = false;

#ifdef DEBUG
    Serial.print(F("Buffer("));
    Serial.print(count);
    Serial.print(F("): "));
    for (int j = 0; j < count; j++) {
      c = data[j];
      if (isprint(c)) {
        Serial.print(c);
      } else {
        Serial.print(c < 0x10 ? F("[0") : F("["));
        Serial.print(c, HEX);
        Serial.print(F("]"));
      }
    }
    Serial.println();
#endif
    // セミコロン (;) までの全文字列を読み取り
    for (int j = 0; j < count && (c = *data++) != ';'; j++) {
      // 次状態判定
      switch (tokenStatus) {
        case INITIAL:
          if (isalpha(c)) {
            nextTokenStatus = IDENTIFIER;
          } else if (isdigit(c) || c == '+' || c == '-') {
            nextTokenStatus = NUMBER;
          } else if (c == ',') {
            nextTokenStatus = COMMA;
          } else if (c == ' ' || c == '\t') {
            nextTokenStatus = INITIAL;
          } else {
            nextTokenStatus = ERROR;
          }
          break;
        case IDENTIFIER:
          if (isalpha(c) || isdigit(c)) {
            nextTokenStatus = IDENTIFIER;
          } else if (c == ' ' || c == '\t' || c == ',' || c == '\r' || c == '\n') {
            nextTokenStatus = COMPLETED;
          } else {
            nextTokenStatus = ERROR;
          }
          break;
        case NUMBER:
          if (isdigit(c) || c == '+' || c == '-') {
            nextTokenStatus = NUMBER;
          } else if (c == ' ' || c == '\t' || c == ',' || c == '\r' || c == '\n') {
            nextTokenStatus = COMPLETED;
          } else {
            nextTokenStatus = ERROR;
          }
          break;
        case COMMA:
          nextTokenStatus = COMPLETED;
          break;
        case COMPLETED:
          nextTokenStatus = COMPLETED;
          break;
        case ERROR:
          if (c == ' ' || c == '\t' || c == ',' || c == '\r' || c == '\n') {
            nextTokenStatus = COMPLETED;
          } else {
            nextTokenStatus = ERROR;
          }
          break;
      }

      if (!(nextTokenStatus == INITIAL
            || nextTokenStatus == COMPLETED)) {
        if (fieldCount < FIELD_LENGTH) {
          field[fieldCount++] = c;
        }
      }
      // トークン末尾到達
      if (nextTokenStatus == COMPLETED) {
        // 文字列終端に達した場合を除き、読み過ぎた文字を戻しておく
        if (c != '\r' && c != '\n') {
          j--; data--;
        }

        if (tokenStatus == ERROR) {
          // 文法エラーを印刷
          if (valid) {
            valid = false;
            if (!valid) {
              Serial.print(F("Error[PT01] at line "));
              Serial.println(this->lineCount);
              Serial.print(F("Syntax error \""));
              for (int i = 0; i < fieldCount && i < FIELD_LENGTH; i++) {
                Serial.print(field[i]);
              }
              Serial.println("\"");
            }
          }
        } else {
          // 文法エラーなき場合、トークン状態
          switch (fieldId) {
            case 0:
              // 第1トークン (識別子or即値) 解析
              if (valid) {
                if (tokenStatus == IDENTIFIER) {
                  valid &= this->parseIdentifier(_t0, field, fieldCount);
                } else if (tokenStatus == NUMBER) {
                  valid &= this->parseNumber(_t0, field, fieldCount);
                }
              }
              break;
            case 1:
              // 第2トークンはカンマであったか
              commaInPlace = tokenStatus == COMMA;
              break;
            case 2:
              // 第3トークン (識別子or即値) 解析
              if (valid) {
                if (tokenStatus == IDENTIFIER) {
                  valid &= this->parseIdentifier(_t1, field, fieldCount);
                } else if (tokenStatus == NUMBER) {
                  valid &= this->parseNumber(_t1, field, fieldCount);
                }
              }
              break;
            default:
              break;
          }
        }
        // フィールドを再初期化
        nextTokenStatus = INITIAL;
        fieldId++;
        fieldCount = 0;
      }
      tokenStatus = nextTokenStatus;
    }
    // 入力フィールド数に応じて構文解析
    switch (fieldId) {
      case 0:
        valid = false;
        break;
      case 1:
        if (_t0.isEmpty() || _t0.isImmediate()) {
          if (valid) {
            valid = false;
            if (!valid) {
              Serial.print(F("Error[PF01] at line "));
              Serial.println(this->lineCount);
              Serial.println(F("Expected COMMAND or INPUT device"));
            }
          }
        } else {
          if (_t0.isCommand()) {
            valid = true;
          } else {
            if (valid) {
              valid = false;
              if (!valid) {
                Serial.print(F("Error[PF03] at line "));
                Serial.println(this->lineCount);
                Serial.println(F("Expected comma (\",\")"));
              }
            }
          }
        }
        break;
      case 2:
        if (_t0.isEmpty() || _t0.isImmediate()) {
          if (valid) {
            valid = false;
            if (!valid) {
              Serial.print(F("Error[PF01] at line "));
              Serial.println(this->lineCount);
              Serial.println(F("Expected COMMAND or INPUT device"));
            }
          }
        } else {
          if (_t0.isCommand()) {
            if (valid) {
              valid = false;
              if (!valid) {
                Serial.print(F("Error[PF02] at line "));
                Serial.println(this->lineCount);
                Serial.println(F("Excess parameter"));
              }
            }
          } else {
            if (commaInPlace) {
              if (valid) {
                valid = false;
                if (!valid) {
                  Serial.print(F("Error[PF04] at line "));
                  Serial.println(this->lineCount);
                  Serial.println(F("Expected OUTPUT device or NUMBER"));
                }
              }
            } else {
              if (valid) {
                valid = false;
                if (!valid) {
                  Serial.print(F("Error[PF03] at line "));
                  Serial.println(this->lineCount);
                  Serial.println(F("Expected comma (\",\")"));
                }
              }
            }
          }
        }
        break;
      case 3:
        if (_t0.isEmpty() || _t0.isImmediate()) {
          if (valid) {
            valid = false;
            if (!valid) {
              Serial.print(F("Error[PF01] at line "));
              Serial.println(this->lineCount);
              Serial.println(F("Expected COMMAND or INPUT device"));
            }
          }
        } else {
          if (_t0.isCommand()) {
            if (valid) {
              valid = false;
              if (!valid) {
                Serial.print(F("Error[PF02] at line "));
                Serial.println(this->lineCount);
                Serial.println(F("Excess parameter"));
              }
            }
          } else {
            if (valid) {
              valid = commaInPlace;
              if (!valid) {
                Serial.print(F("Error[PF03] at line "));
                Serial.println(this->lineCount);
                Serial.println(F("Expected comma (\",\")"));
              }
            }
            if (valid) {
              valid = !_t1.isCommand();
              if (!valid) {
                Serial.print(F("Error[PF04] at line "));
                Serial.println(this->lineCount);
                Serial.println(F("Expected OUTPUT device or NUMBER"));
              }
            }
          }
        }
        break;
      default:
        if (_t0.isImmediate()) {
          if (valid) {
            valid = false;
            if (!valid) {
              Serial.print(F("Error[PF01] at line "));
              Serial.println(this->lineCount);
              Serial.println(F("Expected COMMAND or INPUT device"));
            }
          }
        } else {
          if (valid) {
            valid = false;
            if (!valid) {
              Serial.print(F("Error[PF02] at line "));
              Serial.println(this->lineCount);
              Serial.println(F("Excess parameter"));
            }
          }
        }
        break;
    }
    this->clear();
  } else {
    // バッファ内メッセージ未完結
    valid = false;
  }
  return valid;
}

bool Parser::checkRun() {
  bool result;

  result = this->bufferCompleted
           && strncmp("RUN", this->buffer->getData(), 3) == 0;
  if (result) {
    // バッファ消去はヒット時のみ (このあと構文解析に掛けるので)
    this->clear();
  }
  return result;
}

bool Parser::checkStop() {
  bool result;

  result = this->bufferCompleted
           && (strncmp("STOP", this->buffer->getData(), 4) == 0);
  if (this->bufferCompleted) {
    this->clear();
  }
  return result;
}

