#include <Arduino.h>
#include "RingBuffer.h"

// RingBuffer のコンストラクタ
RingBuffer::RingBuffer(int _length) {
  this->length = 0;
  this->rp = this->wp = 0;
  this->data = NULL;
  this->allocate(_length);
}

// RingBuffer のデストラクタ
RingBuffer::~RingBuffer() {
  this->deallocate();
}

// バッファ領域の確保
void RingBuffer::allocate(int _length) {
  this->data = new char[_length];
  this->length = this->data ? _length : 0;
}

// バッファ領域の解放
void RingBuffer::deallocate() {
  delete[] this->data;
}

// データの有無を確認。1 バイトでもあれば 1 を、なければ 0 を返却
bool RingBuffer::available() {
  return this->full || this->rp != this->wp;
}

// データを全消去
void RingBuffer::clear() {
  this->wp = this->rp;
  this->full = false;
}

// データ 1 バイトを格納。成功時 0 を、失敗時非 0 を返却
int RingBuffer::put(char _c) {
  int result = 0;
  
  if (this->full) {
    result = 1;
  } else {
    this->data[this->wp] = _c;
    // 書き込みポインタを前進
    this->wp = this->wp >= this->length - 1 ? 0 : this->wp + 1;
    // 満杯かどうかチェック
    this->full = this->wp == this->rp;
    result = 0;
  }
  return result;
}

// データ 1 バイトを取得。成功時 0 を、失敗時非 0 を返却
int RingBuffer::get(char& _c) {
  int result = 0;
  
  if (this->available()) {
    _c = this->data[this->rp];
    // 読み取りポインタを前進
    this->rp = this->rp >= this->length - 1 ? 0 : this->rp + 1;
    // 満杯フラグをクリア
    this->full = false;
    result = 0;
  } else {
    result = 1;
  }
  return result;
}

// バッファ領域へのポインタを取得
char* RingBuffer::getData() {
  return this->data;
}

