#include <Arduino.h>
#include "TModule.h"

// カウンタをクリア
void TModule::clear() {
  this->count = 0;
  this->innerCount = 0;
}

// カウンタ値をセット
void TModule::set(int _count) {
  this->count = _count;
  this->innerCount = 0;
}

// ベース時間倍率をセット
void TModule::setMultiplier(int _multiplier) {
  this->multiplier = _multiplier;
  this->innerCount = this->innerCount >= _multiplier ? _multiplier - 1 : this->innerCount;
}

// タイマーモジュールのコンストラクタ
TModule::TModule(int _multiplier) {
  this->multiplier = _multiplier > 0 ? _multiplier : 1;
  this->innerCount = 0;
  this->count = 0;
}

// タイマーを 1 単位進行
bool TModule::step() {
  bool timeout = this->multiplier == 1 ?
                 (this->count == 1) :
                 (this->count == 0 && this->innerCount == 1);
  if (this->innerCount > 0) {
    this->innerCount--;
  } else {
    if (this->count > 0) {
      this->count--;
      this->innerCount = this->multiplier - 1;
    }
  }
  return timeout;
}

// タイマーが走行中であるか返却 (0: 待機中、1: 走行中)
bool TModule::isRunning() {
  return (this->count > 0) || (this->innerCount > 0);
}

