#include <Arduino.h>
#include <limits.h>
#include "Common.h"
#include "TimerCore.h"
#include "Token.h"

// 即値テーブル領域のアドレス原点
int TimerCore::INTEGER_IMM_ADDRESS_BASE = 0x00;

// 即値テーブルサイズ: アドレス範囲 0x00..0x3f の2バイト32組
int TimerCore::NUM_IMM_SLOTS = 32;

TimerCore::TimerCore() {
  this->reset();
  this->clear();
}

TimerCore::~TimerCore() {}

void TimerCore::loadWireCode(unsigned char* blob) {
  unsigned char b;
  unsigned char bLow, bHigh;

  // タイマー立ち上がりエッジトリガ
  for (int i = 0; i < 8; i++) {
    b = blob[(int) TokenType::BI_TIMER_POSEDGE0 + i];
    if (b == 0xff) {
      this->conn.timerPositiveEdge[i].type = ConnType::NOT_CONNECTED;
    } else if (b < TimerCore::NUM_IMM_SLOTS + 2) {
      this->conn.timerPositiveEdge[i].type = ConnType::IMMEDIATE;
      if (b == 0 || b == 1) {
        this->conn.timerPositiveEdge[i].value.number = b;
      } else {
        bLow = blob[INTEGER_IMM_ADDRESS_BASE + (b - 2) * 2];
        bHigh = blob[INTEGER_IMM_ADDRESS_BASE + (b - 2) * 2 + 1];
        this->conn.timerPositiveEdge[i].value.number = ((int) bHigh << 8) + bLow;
      }
    } else {
      this->conn.timerPositiveEdge[i].type = ConnType::DEVICE;
      this->conn.timerPositiveEdge[i].value.device = (TokenType) (b | 0x100);
    }
  }
  // タイマー立ち下がりエッジトリガ
  for (int i = 0; i < 8; i++) {
    b = blob[(int) TokenType::BI_TIMER_NEGEDGE0 + i];
    if (b == 0xff) {
      this->conn.timerNegativeEdge[i].type = ConnType::NOT_CONNECTED;
    } else if (b < TimerCore::NUM_IMM_SLOTS + 2) {
      this->conn.timerNegativeEdge[i].type = ConnType::IMMEDIATE;
      if (b == 0 || b == 1) {
        this->conn.timerNegativeEdge[i].value.number = b;
      } else {
        bLow = blob[INTEGER_IMM_ADDRESS_BASE + (b - 2) * 2];
        bHigh = blob[INTEGER_IMM_ADDRESS_BASE + (b - 2) * 2 + 1];
        this->conn.timerNegativeEdge[i].value.number = ((int) bHigh << 8) + bLow;
      }
    } else {
      this->conn.timerNegativeEdge[i].type = ConnType::DEVICE;
      this->conn.timerNegativeEdge[i].value.device = (TokenType) (b | 0x100);
    }
  }
  // タイマーレベルトリガ
  for (int i = 0; i < 8; i++) {
    b = blob[(int) TokenType::BI_TIMER_LEVELTRIG0 + i];
    if (b == 0xff) {
      this->conn.timerLevelTrigger[i].type = ConnType::NOT_CONNECTED;
    } else if (b < TimerCore::NUM_IMM_SLOTS + 2) {
      this->conn.timerLevelTrigger[i].type = ConnType::IMMEDIATE;
      if (b == 0 || b == 1) {
        this->conn.timerLevelTrigger[i].value.number = b;
      } else {
        bLow = blob[INTEGER_IMM_ADDRESS_BASE + (b - 2) * 2];
        bHigh = blob[INTEGER_IMM_ADDRESS_BASE + (b - 2) * 2 + 1];
        this->conn.timerLevelTrigger[i].value.number = ((int) bHigh << 8) + bLow;
      }
    } else {
      this->conn.timerLevelTrigger[i].type = ConnType::DEVICE;
      this->conn.timerLevelTrigger[i].value.device = (TokenType) (b | 0x100);
    }
  }
  // タイマーリセット
  for (int i = 0; i < 8; i++) {
    b = blob[(int) TokenType::BI_TIMER_RESET0 + i];
    if (b == 0xff) {
      this->conn.timerReset[i].type = ConnType::NOT_CONNECTED;
    } else if (b < TimerCore::NUM_IMM_SLOTS + 2) {
      this->conn.timerReset[i].type = ConnType::IMMEDIATE;
      if (b == 0 || b == 1) {
        this->conn.timerReset[i].value.number = b;
      } else {
        bLow = blob[INTEGER_IMM_ADDRESS_BASE + (b - 2) * 2];
        bHigh = blob[INTEGER_IMM_ADDRESS_BASE + (b - 2) * 2 + 1];
        this->conn.timerReset[i].value.number = ((int) bHigh << 8) + bLow;
      }
    } else {
      this->conn.timerReset[i].type = ConnType::DEVICE;
      this->conn.timerReset[i].value.device = (TokenType) (b | 0x100);
    }
  }
  // 論理 AND ゲート
  for (int i = 0; i < 4; i++) {
    for (int j = 0; j < 8; j++) {
      b = blob[(int) TokenType::BI_ANDGATE_INA0 + j * 4 + i];
      if (b == 0xff) {
        this->conn.logicAndGate[i][j].type = ConnType::NOT_CONNECTED;
      } else if (b < TimerCore::NUM_IMM_SLOTS + 2) {
        this->conn.logicAndGate[i][j].type = ConnType::IMMEDIATE;
        if (b == 0 || b == 1) {
          this->conn.logicAndGate[i][j].value.number = b;
        } else {
          bLow = blob[INTEGER_IMM_ADDRESS_BASE + (b - 2) * 2];
          bHigh = blob[INTEGER_IMM_ADDRESS_BASE + (b - 2) * 2 + 1];
          this->conn.logicAndGate[i][j].value.number = ((int) bHigh << 8) + bLow;
        }
      } else {
        this->conn.logicAndGate[i][j].type = ConnType::DEVICE;
        this->conn.logicAndGate[i][j].value.device = (TokenType) (b | 0x100);
      }
    }
  }
  // 論理 OR ゲート
  for (int i = 0; i < 4; i++) {
    for (int j = 0; j < 8; j++) {
      b = blob[(int) TokenType::BI_ORGATE_INA0 + j * 4 + i];
      if (b == 0xff) {
        this->conn.logicOrGate[i][j].type = ConnType::NOT_CONNECTED;
      } else if (b < TimerCore::NUM_IMM_SLOTS + 2) {
        this->conn.logicOrGate[i][j].type = ConnType::IMMEDIATE;
        if (b == 0 || b == 1) {
          this->conn.logicOrGate[i][j].value.number = b;
        } else {
          bLow = blob[INTEGER_IMM_ADDRESS_BASE + (b - 2) * 2];
          bHigh = blob[INTEGER_IMM_ADDRESS_BASE + (b - 2) * 2 + 1];
          this->conn.logicOrGate[i][j].value.number = ((int) bHigh << 8) + bLow;
        }
      } else {
        this->conn.logicOrGate[i][j].type = ConnType::DEVICE;
        this->conn.logicOrGate[i][j].value.device = (TokenType) (b | 0x100);
      }
    }
  }
  // 論理 NOT ゲート
  for (int i = 0; i < 8; i++) {
    b = blob[(int) TokenType::BI_NOTGATE_IN0 + i];
    if (b == 0xff) {
      this->conn.logicInverter[i].type = ConnType::NOT_CONNECTED;
    } else if (b < TimerCore::NUM_IMM_SLOTS + 2) {
      this->conn.logicInverter[i].type = ConnType::IMMEDIATE;
      if (b == 0 || b == 1) {
        this->conn.logicInverter[i].value.number = b;
      } else {
        bLow = blob[INTEGER_IMM_ADDRESS_BASE + (b - 2) * 2];
        bHigh = blob[INTEGER_IMM_ADDRESS_BASE + (b - 2) * 2 + 1];
        this->conn.logicInverter[i].value.number = ((int) bHigh << 8) + bLow;
      }
    } else {
      this->conn.logicInverter[i].type = ConnType::DEVICE;
      this->conn.logicInverter[i].value.device = (TokenType) (b | 0x100);
    }
  }
  // リレー
  for (int i = 0; i < 4; i++) {
    b = blob[(int) TokenType::BI_RELAY0 + i];
    if (b == 0xff) {
      this->conn.relay[i].type = ConnType::NOT_CONNECTED;
    } else if (b < TimerCore::NUM_IMM_SLOTS + 2) {
      this->conn.relay[i].type = ConnType::IMMEDIATE;
      if (b == 0 || b == 1) {
        this->conn.relay[i].value.number = b;
      } else {
        bLow = blob[INTEGER_IMM_ADDRESS_BASE + (b - 2) * 2];
        bHigh = blob[INTEGER_IMM_ADDRESS_BASE + (b - 2) * 2 + 1];
        this->conn.relay[i].value.number = ((int) bHigh << 8) + bLow;
      }
    } else {
      this->conn.relay[i].type = ConnType::DEVICE;
      this->conn.relay[i].value.device = (TokenType) (b | 0x100);
    }
  }
  // ジャンパー選択入力 (L レベル)
  for (int i = 0; i < 6; i++) {
    b = blob[(int) TokenType::II_JP_MUXL0 + i];
    if (b == 0xff) {
      this->conn.jumperMuxL[i].type = ConnType::NOT_CONNECTED;
    } else if (b < TimerCore::NUM_IMM_SLOTS + 2) {
      this->conn.jumperMuxL[i].type = ConnType::IMMEDIATE;
      if (b == 0 || b == 1) {
        this->conn.jumperMuxL[i].value.number = b;
      } else {
        bLow = blob[INTEGER_IMM_ADDRESS_BASE + (b - 2) * 2];
        bHigh = blob[INTEGER_IMM_ADDRESS_BASE + (b - 2) * 2 + 1];
        this->conn.jumperMuxL[i].value.number = ((int) bHigh << 8) + bLow;
      }
    } else {
      this->conn.jumperMuxL[i].type = ConnType::DEVICE;
      this->conn.jumperMuxL[i].value.device = (TokenType) (b | 0x100);
    }
  }
  // ジャンパー選択入力 (H レベル)
  for (int i = 0; i < 6; i++) {
    b = blob[(int) TokenType::II_JP_MUXH0 + i];
    if (b == 0xff) {
      this->conn.jumperMuxH[i].type = ConnType::NOT_CONNECTED;
    } else if (b < TimerCore::NUM_IMM_SLOTS + 2) {
      this->conn.jumperMuxH[i].type = ConnType::IMMEDIATE;
      if (b == 0 || b == 1) {
        this->conn.jumperMuxH[i].value.number = b;
      } else {
        bLow = blob[INTEGER_IMM_ADDRESS_BASE + (b - 2) * 2];
        bHigh = blob[INTEGER_IMM_ADDRESS_BASE + (b - 2) * 2 + 1];
        this->conn.jumperMuxH[i].value.number = ((int) bHigh << 8) + bLow;
      }
    } else {
      this->conn.jumperMuxH[i].type = ConnType::DEVICE;
      this->conn.jumperMuxH[i].value.device = (TokenType) (b | 0x100);
    }
  }
  // タイマーカウンター設定
  for (int i = 0; i < 8; i++) {
    b = blob[(int) TokenType::II_TIMER_CNT0 + i];
    if (b == 0xff) {
      this->conn.timerCounter[i].type = ConnType::NOT_CONNECTED;
    } else if (b < TimerCore::NUM_IMM_SLOTS + 2) {
      this->conn.timerCounter[i].type = ConnType::IMMEDIATE;
      if (b == 0 || b == 1) {
        this->conn.timerCounter[i].value.number = b;
      } else {
        bLow = blob[INTEGER_IMM_ADDRESS_BASE + (b - 2) * 2];
        bHigh = blob[INTEGER_IMM_ADDRESS_BASE + (b - 2) * 2 + 1];
        this->conn.timerCounter[i].value.number = ((int) bHigh << 8) + bLow;
      }
    } else {
      this->conn.timerCounter[i].type = ConnType::DEVICE;
      this->conn.timerCounter[i].value.device = (TokenType) (b | 0x100);
    }
  }
  // タイマー倍率設定
  for (int i = 0; i < 8; i++) {
    b = blob[(int) TokenType::II_TIMER_MULT0 + i];
    if (b == 0xff) {
      this->conn.timerMultiplier[i].type = ConnType::NOT_CONNECTED;
    } else if (b < TimerCore::NUM_IMM_SLOTS + 2) {
      this->conn.timerMultiplier[i].type = ConnType::IMMEDIATE;
      if (b == 0 || b == 1) {
        this->conn.timerMultiplier[i].value.number = b;
      } else {
        bLow = blob[INTEGER_IMM_ADDRESS_BASE + (b - 2) * 2];
        bHigh = blob[INTEGER_IMM_ADDRESS_BASE + (b - 2) * 2 + 1];
        this->conn.timerMultiplier[i].value.number = ((int) bHigh << 8) + bLow;
      }
    } else {
      this->conn.timerMultiplier[i].type = ConnType::DEVICE;
      this->conn.timerMultiplier[i].value.device = (TokenType) (b | 0x100);
    }
  }
  // 加算器
  for (int i = 0; i < 4; i++) {
    for (int j = 0; j < 2; j++) {
      b = blob[(int) TokenType::II_ADDER_INA0 + j * 4 + i];
      if (b == 0xff) {
        this->conn.adder[i][j].type = ConnType::NOT_CONNECTED;
      } else if (b < TimerCore::NUM_IMM_SLOTS + 2) {
        this->conn.adder[i][j].type = ConnType::IMMEDIATE;
        if (b == 0 || b == 1) {
          this->conn.adder[i][j].value.number = b;
        } else {
          bLow = blob[INTEGER_IMM_ADDRESS_BASE + (b - 2) * 2];
          bHigh = blob[INTEGER_IMM_ADDRESS_BASE + (b - 2) * 2 + 1];
          this->conn.adder[i][j].value.number = ((int) bHigh << 8) + bLow;
        }
      } else {
        this->conn.adder[i][j].type = ConnType::DEVICE;
        this->conn.adder[i][j].value.device = (TokenType) (b | 0x100);
      }
    }
  }
  // 乗算器
  for (int i = 0; i < 4; i++) {
    for (int j = 0; j < 2; j++) {
      b = blob[(int) TokenType::II_MULTIPLIER_INA0 + j * 4 + i];
      if (b == 0xff) {
        this->conn.multiplier[i][j].type = ConnType::NOT_CONNECTED;
      } else if (b < TimerCore::NUM_IMM_SLOTS + 2) {
        this->conn.multiplier[i][j].type = ConnType::IMMEDIATE;
        if (b == 0 || b == 1) {
          this->conn.multiplier[i][j].value.number = b;
        } else {
          bLow = blob[INTEGER_IMM_ADDRESS_BASE + (b - 2) * 2];
          bHigh = blob[INTEGER_IMM_ADDRESS_BASE + (b - 2) * 2 + 1];
          this->conn.multiplier[i][j].value.number = ((int) bHigh << 8) + bLow;
        }
      } else {
        this->conn.multiplier[i][j].type = ConnType::DEVICE;
        this->conn.multiplier[i][j].value.device = (TokenType) (b | 0x100);
      }
    }
  }
  // 半固定抵抗最大設定
  for (int i = 0; i < 4; i++) {
    b = blob[(int) TokenType::II_VR_MAX0 + i];
    if (b == 0xff) {
      this->conn.potmeterMax[i].type = ConnType::NOT_CONNECTED;
    } else if (b < TimerCore::NUM_IMM_SLOTS + 2) {
      this->conn.potmeterMax[i].type = ConnType::IMMEDIATE;
      if (b == 0 || b == 1) {
        this->conn.potmeterMax[i].value.number = b;
      } else {
        bLow = blob[INTEGER_IMM_ADDRESS_BASE + (b - 2) * 2];
        bHigh = blob[INTEGER_IMM_ADDRESS_BASE + (b - 2) * 2 + 1];
        this->conn.potmeterMax[i].value.number = ((int) bHigh << 8) + bLow;
      }
    } else {
      this->conn.potmeterMax[i].type = ConnType::DEVICE;
      this->conn.potmeterMax[i].value.device = (TokenType) (b | 0x100);
    }
  }
}

// 接続元 _c からのバイナリ出力の値を取得し、_x に代入。無接続時は -1 を返却 (_x の結果は無効)
inline int TimerCore::getBinaryOutValue(bool& _x, Connection _c) {
  int result = 0;

  if (_c.type == ConnType::NOT_CONNECTED) {
    result = -1;
  } else if (_c.type == ConnType::IMMEDIATE) {
    _x = _c.value.number != 0;
  } else {
    switch (_c.value.device) {
      case TokenType::BO_CONST_DELAY:
        _x = this->tickCount >= 10L;
        break;
      case TokenType::BO_CONST_ALWAYS:
        _x = 1;
        break;
      case TokenType::BO_PHOTO0:
      case TokenType::BO_PHOTO1:
      case TokenType::BO_PHOTO2:
      case TokenType::BO_PHOTO3:
        _x = this->reg.photoCoupler
             & (0b1 << ((int) _c.value.device - (int) TokenType::BO_PHOTO0));
        break;
      case TokenType::BO_JP_LEVEL0:
      case TokenType::BO_JP_LEVEL1:
      case TokenType::BO_JP_LEVEL2:
      case TokenType::BO_JP_LEVEL3:
      case TokenType::BO_JP_LEVEL4:
      case TokenType::BO_JP_LEVEL5:
        _x = this->reg.jumperLevel
             & (0b1 << ((int) _c.value.device - (int) TokenType::BO_JP_LEVEL0));
        break;
      case TokenType::BO_TIMER_OUT0:
      case TokenType::BO_TIMER_OUT1:
      case TokenType::BO_TIMER_OUT2:
      case TokenType::BO_TIMER_OUT3:
      case TokenType::BO_TIMER_OUT4:
      case TokenType::BO_TIMER_OUT5:
      case TokenType::BO_TIMER_OUT6:
      case TokenType::BO_TIMER_OUT7:
        _x = this->reg.timerOut
             & (0b1 << ((int) _c.value.device - (int) TokenType::BO_TIMER_OUT0));
        break;
      case TokenType::BO_TIMER_DELAY0:
      case TokenType::BO_TIMER_DELAY1:
      case TokenType::BO_TIMER_DELAY2:
      case TokenType::BO_TIMER_DELAY3:
      case TokenType::BO_TIMER_DELAY4:
      case TokenType::BO_TIMER_DELAY5:
      case TokenType::BO_TIMER_DELAY6:
      case TokenType::BO_TIMER_DELAY7:
        _x = this->reg.timerDelay
             & (0b1 << ((int) _c.value.device - (int) TokenType::BO_TIMER_DELAY0));
        break;
      case TokenType::BO_ANDGATE_OUT0:
      case TokenType::BO_ANDGATE_OUT1:
      case TokenType::BO_ANDGATE_OUT2:
      case TokenType::BO_ANDGATE_OUT3:
        _x = this->reg.logicAndGate
             & (0b1 << ((int) _c.value.device - (int) TokenType::BO_ANDGATE_OUT0));
        break;
      case TokenType::BO_ORGATE_OUT0:
      case TokenType::BO_ORGATE_OUT1:
      case TokenType::BO_ORGATE_OUT2:
      case TokenType::BO_ORGATE_OUT3:
        _x = this->reg.logicOrGate
             & (0b1 << ((int) _c.value.device - (int) TokenType::BO_ORGATE_OUT0));
        break;
      case TokenType::BO_NOTGATE_OUT0:
      case TokenType::BO_NOTGATE_OUT1:
      case TokenType::BO_NOTGATE_OUT2:
      case TokenType::BO_NOTGATE_OUT3:
      case TokenType::BO_NOTGATE_OUT4:
      case TokenType::BO_NOTGATE_OUT5:
      case TokenType::BO_NOTGATE_OUT6:
      case TokenType::BO_NOTGATE_OUT7:
        _x = this->reg.logicInverter
             & (0b1 << ((int) _c.value.device - (int) TokenType::BO_NOTGATE_OUT0));
        break;
      case TokenType::IO_JP_MUXOUT0:
      case TokenType::IO_JP_MUXOUT1:
      case TokenType::IO_JP_MUXOUT2:
      case TokenType::IO_JP_MUXOUT3:
      case TokenType::IO_JP_MUXOUT4:
      case TokenType::IO_JP_MUXOUT5:
        _x = !!(this->reg.jumperMux[(int) _c.value.device - (int) TokenType::IO_JP_MUXOUT0]);
        break;
      case TokenType::IO_ADDER_OUT0:
      case TokenType::IO_ADDER_OUT1:
      case TokenType::IO_ADDER_OUT2:
      case TokenType::IO_ADDER_OUT3:
        _x = !!(this->reg.adder[(int) _c.value.device - (int) TokenType::IO_ADDER_OUT0]);
        break;
      case TokenType::IO_MULTIPLIER_OUT0:
      case TokenType::IO_MULTIPLIER_OUT1:
      case TokenType::IO_MULTIPLIER_OUT2:
      case TokenType::IO_MULTIPLIER_OUT3:
        _x = !!(this->reg.multiplier[(int) _c.value.device - (int) TokenType::IO_MULTIPLIER_OUT0]);
        break;
      case TokenType::IO_VR_OUT0:
      case TokenType::IO_VR_OUT1:
      case TokenType::IO_VR_OUT2:
      case TokenType::IO_VR_OUT3:
        _x = !!(this->reg.potmeterRanged[(int) _c.value.device - (int) TokenType::IO_VR_OUT0]);
        break;
      default:
        result = -1;
        break;
    }
  }
  return result;
}

// 接続元 _c からの整数出力の値を取得し、_x に代入。無接続時は -1 を返却 (_x の結果は無効)
inline int TimerCore::getIntegerOutValue(int& _x, Connection _c) {
  int result = 0;

  if (_c.type == ConnType::NOT_CONNECTED) {
    result = -1;
  } else if (_c.type == ConnType::IMMEDIATE) {
    _x = _c.value.number;
  } else {
    switch (_c.value.device) {
      case TokenType::BO_CONST_DELAY:
        _x = this->tickCount >= 10L;
        break;
      case TokenType::BO_CONST_ALWAYS:
        _x = 1;
        break;
      case TokenType::BO_PHOTO0:
      case TokenType::BO_PHOTO1:
      case TokenType::BO_PHOTO2:
      case TokenType::BO_PHOTO3:
        _x = this->reg.photoCoupler
             & (0b1 << ((int) _c.value.device - (int) TokenType::BO_PHOTO0));
        break;
      case TokenType::BO_JP_LEVEL0:
      case TokenType::BO_JP_LEVEL1:
      case TokenType::BO_JP_LEVEL2:
      case TokenType::BO_JP_LEVEL3:
      case TokenType::BO_JP_LEVEL4:
      case TokenType::BO_JP_LEVEL5:
        _x = this->reg.jumperLevel
             & (0b1 << ((int) _c.value.device - (int) TokenType::BO_JP_LEVEL0));
        break;
      case TokenType::BO_TIMER_OUT0:
      case TokenType::BO_TIMER_OUT1:
      case TokenType::BO_TIMER_OUT2:
      case TokenType::BO_TIMER_OUT3:
      case TokenType::BO_TIMER_OUT4:
      case TokenType::BO_TIMER_OUT5:
      case TokenType::BO_TIMER_OUT6:
      case TokenType::BO_TIMER_OUT7:
        _x = this->reg.timerOut
             & (0b1 << ((int) _c.value.device - (int) TokenType::BO_TIMER_OUT0));
        break;
      case TokenType::BO_TIMER_DELAY0:
      case TokenType::BO_TIMER_DELAY1:
      case TokenType::BO_TIMER_DELAY2:
      case TokenType::BO_TIMER_DELAY3:
      case TokenType::BO_TIMER_DELAY4:
      case TokenType::BO_TIMER_DELAY5:
      case TokenType::BO_TIMER_DELAY6:
      case TokenType::BO_TIMER_DELAY7:
        _x = this->reg.timerDelay
             & (0b1 << ((int) _c.value.device - (int) TokenType::BO_TIMER_DELAY0));
        break;
      case TokenType::BO_ANDGATE_OUT0:
      case TokenType::BO_ANDGATE_OUT1:
      case TokenType::BO_ANDGATE_OUT2:
      case TokenType::BO_ANDGATE_OUT3:
        _x = this->reg.logicAndGate
             & (0b1 << ((int) _c.value.device - (int) TokenType::BO_ANDGATE_OUT0));
        break;
      case TokenType::BO_ORGATE_OUT0:
      case TokenType::BO_ORGATE_OUT1:
      case TokenType::BO_ORGATE_OUT2:
      case TokenType::BO_ORGATE_OUT3:
        _x = this->reg.logicOrGate
             & (0b1 << ((int) _c.value.device - (int) TokenType::BO_ORGATE_OUT0));
        break;
      case TokenType::BO_NOTGATE_OUT0:
      case TokenType::BO_NOTGATE_OUT1:
      case TokenType::BO_NOTGATE_OUT2:
      case TokenType::BO_NOTGATE_OUT3:
      case TokenType::BO_NOTGATE_OUT4:
      case TokenType::BO_NOTGATE_OUT5:
      case TokenType::BO_NOTGATE_OUT6:
      case TokenType::BO_NOTGATE_OUT7:
        _x = this->reg.logicInverter
             & (0b1 << ((int) _c.value.device - (int) TokenType::BO_NOTGATE_OUT0));
        break;
      case TokenType::IO_JP_MUXOUT0:
      case TokenType::IO_JP_MUXOUT1:
      case TokenType::IO_JP_MUXOUT2:
      case TokenType::IO_JP_MUXOUT3:
      case TokenType::IO_JP_MUXOUT4:
      case TokenType::IO_JP_MUXOUT5:
        _x = this->reg.jumperMux[(int) _c.value.device - (int) TokenType::IO_JP_MUXOUT0];
        break;
      case TokenType::IO_ADDER_OUT0:
      case TokenType::IO_ADDER_OUT1:
      case TokenType::IO_ADDER_OUT2:
      case TokenType::IO_ADDER_OUT3:
        _x = this->reg.adder[(int) _c.value.device - (int) TokenType::IO_ADDER_OUT0];
        break;
      case TokenType::IO_MULTIPLIER_OUT0:
      case TokenType::IO_MULTIPLIER_OUT1:
      case TokenType::IO_MULTIPLIER_OUT2:
      case TokenType::IO_MULTIPLIER_OUT3:
        _x = this->reg.multiplier[(int) _c.value.device - (int) TokenType::IO_MULTIPLIER_OUT0];
        break;
      case TokenType::IO_VR_OUT0:
      case TokenType::IO_VR_OUT1:
      case TokenType::IO_VR_OUT2:
      case TokenType::IO_VR_OUT3:
        _x = this->reg.potmeterRanged[(int) _c.value.device - (int) TokenType::IO_VR_OUT0];
        break;
      default:
        result = -1;
        break;
    }
  }
  return result;
}

void TimerCore::reset() {
  this->tickCount = 0L;
}

void TimerCore::clear() {
  this->reg.jumperLevel = 0b0;
  this->reg.logicAndGate = 0b0;
  this->reg.logicOrGate = 0b0;
  this->reg.logicInverter = 0b0;
  this->reg.timerOut = 0b0;
  this->reg.timerDelay = 0b0;
  this->reg.timerPositiveEdgePrev = 0b0;
  this->reg.timerNegativeEdgePrev = 0b0;
  for (int i = 0; i < 6; i++) {
    this->reg.jumperMux[i] = 0;
  }
  for (int i = 0; i < 4; i++) {
    this->reg.adder[i] = 0;
  }
  for (int i = 0; i < 4; i++) {
    this->reg.multiplier[i] = 0;
  }
  for (int i = 0; i < 4; i++) {
    this->reg.potmeterRanged[i] = 0;
  }
  for (int i = 0; i < 8; i++) {
    this->tm[i].clear();
  }
}

void TimerCore::readFromInputDevices() {
  unsigned char c;

  // フォトカプラ
  c = 0b0;
  for (int i = 0; i < 4; i++) {
    c |= digitalRead(PIN_PHOTOCOUPLER[i]) ? 0b0 : (0b1 << i);
  }
  this->reg.photoCoupler = c;
  // ジャンパーレベル
  c = 0b0;
  for (int i = 0; i < 6; i++) {
    c |= digitalRead(PIN_JUMPER[i]) ? (0b1 << i) : 0b0;
  }
  this->reg.jumperLevel = c;
  // 半固定抵抗 (生の ADC 値)
  for (int i = 0; i < 4; i++) {
    this->reg.potmeterRaw[i] = analogRead(PIN_POTMETER[i]);
  }
}

void TimerCore::writeToOutputDevices() {
  bool b;
  unsigned char c;

  // リレー
  c = 0b0;
  for (int i = 0; i < 4; i++) {
    if (this->getBinaryOutValue(b, this->conn.relay[i]) == -1) {
      b = 0;
    }
    c |= (b << i);
  }
  this->in.relay = c;
  for (int i = 0; i < 4; i++) {
    digitalWrite(PIN_RELAY[i], this->in.relay & (0b1 << i));
  }
}

void TimerCore::iterate() {
  bool b;
  int n0, n1;
  unsigned char c, cj;

  // 論理 NOT ゲート
  c = 0b0;
  for (int i = 0; i < 8; i++) {
    if (this->getBinaryOutValue(b, this->conn.logicInverter[i]) == -1) {
      b = 0;
    }
    c |= (b << i);
  }
  this->reg.logicInverter = ~c;
  // 論理 AND ゲート
  c = 0b0;
  for (int i = 0; i < 4; i++) {
    cj = 0b0;
    for (int j = 0; j < 8; j++) {
      if (this->getBinaryOutValue(b, this->conn.logicAndGate[i][j]) == -1) {
        b = 1;
      }
      cj |= (b << j);
    }
    c |= ((cj == 0b11111111 ? 0b1 : 0b0) << i);
  }
  this->reg.logicAndGate = c;
  // 論理 OR ゲート
  c = 0b0;
  for (int i = 0; i < 4; i++) {
    cj = 0b0;
    for (int j = 0; j < 8; j++) {
      if (this->getBinaryOutValue(b, this->conn.logicOrGate[i][j]) == -1) {
        b = 0;
      }
      cj |= (b << j);
    }
    c |= ((cj != 0b00000000 ? 0b1 : 0b0) << i);
  }
  this->reg.logicOrGate = c;
  // 半固定抵抗
  for (int i = 0; i < 4; i++) {
    // 最大値
    if (this->getIntegerOutValue(n0, this->conn.potmeterMax[i]) == -1) {
      n0 = POTMETER_DEFAULT_VALUE;
    }
    if (n0 < 0) {
      n0 = POTMETER_DEFAULT_VALUE;
    }
    // 生の値と最大値から、範囲補正済みの値を計算
    n1 = (int) (((long) this->reg.potmeterRaw[i] * n0) >> 10);
    this->reg.potmeterRanged[i] = n1;
  }
  // 乗算器
  for (int i = 0; i < 4; i++) {
    if (this->getIntegerOutValue(n0, this->conn.multiplier[i][0]) == -1) {
      n0 = 0;
    }
    if (this->getIntegerOutValue(n1, this->conn.multiplier[i][1]) == -1) {
      n1 = 0;
    }
    this->reg.multiplier[i] = n0 * n1;
  }
  // 加算器
  for (int i = 0; i < 4; i++) {
    if (this->getIntegerOutValue(n0, this->conn.adder[i][0]) == -1) {
      n0 = 0;
    }
    if (this->getIntegerOutValue(n1, this->conn.adder[i][1]) == -1) {
      n1 = 0;
    }
    this->reg.adder[i] = n0 + n1;
  }
  // ジャンパー選択
  for (int i = 0; i < 6; i++) {
    if (this->reg.jumperLevel & (0b1 << i)) {
      // H レベルが選択された
      if (this->getIntegerOutValue(n0, this->conn.jumperMuxH[i]) == -1) {
        n0 = 0;
      }
    } else {
      // L レベルが選択された
      if (this->getIntegerOutValue(n0, this->conn.jumperMuxL[i]) == -1) {
        n0 = 0;
      }
    }
    this->reg.jumperMux[i] = n0;
  }
}

void TimerCore::stepTModules() {
  bool b;
  unsigned char c, cp, cn, timeout;
  int n0, n1;

  // タイマー進行
  timeout = 0b0;
  for (int i = 0; i < 8; i++) {
    if (this->tm[i].step()) {
      timeout |= (0b1 << i);
    }
  }
  // タイマー入力条件設定
  cp = 0b0;
  cn = 0b0;
  for (int i = 0; i < 8; i++) {
    // カウンター設定値 (下限: 0)
    if (this->getIntegerOutValue(n0, this->conn.timerCounter[i]) == -1) {
      n0 = 0;
    }
    if (n0 < 0) {
      n0 = 0;
    }
    // カウンター倍率値 (下限: 1、上限: TIMER_MULT_MAX)
    if (this->getIntegerOutValue(n1, this->conn.timerMultiplier[i]) == -1) {
      n1 = 1;
    }
    if (n1 < 1) {
      n1 = 1;
    }
    if (n1 > TIMER_MULT_MAX) {
      n1 = TIMER_MULT_MAX;
    }
    // 立ち上がりエッジトリガ判定
    if (this->getBinaryOutValue(b, this->conn.timerPositiveEdge[i]) == -1) {
      b = 0;
    }
    cp |= (b << i);
    if (!(this->reg.timerPositiveEdgePrev & (0b1 << i)) && (b == 1)) {
      this->tm[i].set(n0);
      this->tm[i].setMultiplier(n1);
    }
    // 立ち下がりエッジトリガ判定
    if (this->getBinaryOutValue(b, this->conn.timerNegativeEdge[i]) == -1) {
      b = 0;
    }
    cn |= (b << i);
    if ((this->reg.timerNegativeEdgePrev & (0b1 << i)) && (b == 0)) {
      this->tm[i].set(n0);
      this->tm[i].setMultiplier(n1);
    }
    // レベルトリガ判定
    if (this->getBinaryOutValue(b, this->conn.timerLevelTrigger[i]) == -1) {
      b = 0;
    }
    if (b == 1) {
      this->tm[i].set(n0);
      this->tm[i].setMultiplier(n1);
    }
    // レベルリセット判定
    if (this->getBinaryOutValue(b, this->conn.timerReset[i]) == -1) {
      b = 0;
    }
    if (b == 1) {
      this->tm[i].set(0);
    }
  }
  this->reg.timerPositiveEdgePrev = cp;  // 今回の立ち上がりエッジ出力を待避
  this->reg.timerNegativeEdgePrev = cn;  // 今回の立ち下がりエッジ出力を待避
  // 出力をセット
  c = 0b0;
  for (int i = 0; i < 8; i++) {
    if (this->tm[i].isRunning()) {
      c |= (0b1 << i);
    }
  }
  this->reg.timerOut = c;
  this->reg.timerDelay = c | timeout;
}

void TimerCore::step() {
  this->readFromInputDevices();
  for (int i = 0; i < ITERATION_PASSES; i++) {
    this->iterate();
  }
  this->stepTModules();
  this->writeToOutputDevices();
  if (this->tickCount < LONG_MAX) this->tickCount++;
}

