#ifndef TIMERCORE_H_
#define TIMERCORE_H_

#include "Token.h"
#include "TModule.h"

enum class ConnType {
  NOT_CONNECTED = 0,
  IMMEDIATE = 1,
  DEVICE = 2,
};

class Connection {
  public:
    ConnType type;
    union {
      int number;
      TokenType device;
    } value;
};

class TimerCore {
  private:
    long tickCount;
    struct {
      unsigned char relay;  // <3..0>: D..A 出力ビット
    } in;
    struct {
      unsigned char photoCoupler;           // <3..0>: D..A 出力ビット
      unsigned char jumperLevel;            // <5..0>: F..A 出力ビット
      unsigned char timerOut;               // <7..0>: H..A 出力ビット
      unsigned char timerDelay;             // <7..0>: H..A 出力ビット
      unsigned char logicAndGate;           // <3..0>: D..A 出力ビット
      unsigned char logicOrGate;            // <3..0>: D..A 出力ビット
      unsigned char logicInverter;          // <7..0>: H..A 出力ビット
      unsigned char timerPositiveEdgePrev;  // <7..0>: H..A 出力ビット
      unsigned char timerNegativeEdgePrev;  // <7..0>: H..A 出力ビット
      int potmeterRaw[4], potmeterRanged[4];
      int adder[4];
      int multiplier[4];
      int jumperMux[6];
    } reg;
    struct {
      Connection timerPositiveEdge[8];
      Connection timerNegativeEdge[8];
      Connection timerLevelTrigger[8];
      Connection timerReset[8];
      Connection logicAndGate[4][8];
      Connection logicOrGate[4][8];
      Connection logicInverter[8];
      Connection relay[4];
      Connection jumperMuxL[6];
      Connection jumperMuxH[6];
      Connection timerCounter[8];
      Connection timerMultiplier[8];
      Connection adder[4][2];
      Connection multiplier[4][2];
      Connection potmeterMax[4];
    } conn;
    TModule tm[8];
    
  public:
    static int INTEGER_IMM_ADDRESS_BASE;
    static int NUM_IMM_SLOTS;

  private:
    int getBinaryOutValue(bool& _x, Connection _c);
    int getIntegerOutValue(int& _x, Connection _c);
    void readFromInputDevices();
    void writeToOutputDevices();
    void iterate();
    void stepTModules();

  public:
    TimerCore();
    ~TimerCore();
    void reset();
    void clear();
    void step();
    void loadWireCode(unsigned char* _blob);
};

#endif
